
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include <config.h>

#include "soup-enum-types.h"
#include "soup.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)
/* enumerations from "soup-cache.h" */
GType
soup_cacheability_get_type (void)
{
  static gsize id = 0;
  static const GFlagsValue values[] = {
    { C_FLAGS(SOUP_CACHE_CACHEABLE), "SOUP_CACHE_CACHEABLE", "cacheable" },
    { C_FLAGS(SOUP_CACHE_UNCACHEABLE), "SOUP_CACHE_UNCACHEABLE", "uncacheable" },
    { C_FLAGS(SOUP_CACHE_INVALIDATES), "SOUP_CACHE_INVALIDATES", "invalidates" },
    { C_FLAGS(SOUP_CACHE_VALIDATES), "SOUP_CACHE_VALIDATES", "validates" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_flags_register_static ("SoupCacheability", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
GType
soup_cache_type_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_CACHE_SINGLE_USER), "SOUP_CACHE_SINGLE_USER", "single-user" },
    { C_ENUM(SOUP_CACHE_SHARED), "SOUP_CACHE_SHARED", "shared" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupCacheType", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "soup-cookie-jar.h" */
GType
soup_cookie_jar_accept_policy_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_COOKIE_JAR_ACCEPT_ALWAYS), "SOUP_COOKIE_JAR_ACCEPT_ALWAYS", "always" },
    { C_ENUM(SOUP_COOKIE_JAR_ACCEPT_NEVER), "SOUP_COOKIE_JAR_ACCEPT_NEVER", "never" },
    { C_ENUM(SOUP_COOKIE_JAR_ACCEPT_NO_THIRD_PARTY), "SOUP_COOKIE_JAR_ACCEPT_NO_THIRD_PARTY", "no-third-party" },
    { C_ENUM(SOUP_COOKIE_JAR_ACCEPT_GRANDFATHERED_THIRD_PARTY), "SOUP_COOKIE_JAR_ACCEPT_GRANDFATHERED_THIRD_PARTY", "grandfathered-third-party" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupCookieJarAcceptPolicy", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "soup-cookie.h" */
GType
soup_same_site_policy_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_SAME_SITE_POLICY_NONE), "SOUP_SAME_SITE_POLICY_NONE", "none" },
    { C_ENUM(SOUP_SAME_SITE_POLICY_LAX), "SOUP_SAME_SITE_POLICY_LAX", "lax" },
    { C_ENUM(SOUP_SAME_SITE_POLICY_STRICT), "SOUP_SAME_SITE_POLICY_STRICT", "strict" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupSameSitePolicy", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "soup-message-body.h" */
GType
soup_memory_use_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_MEMORY_STATIC), "SOUP_MEMORY_STATIC", "static" },
    { C_ENUM(SOUP_MEMORY_TAKE), "SOUP_MEMORY_TAKE", "take" },
    { C_ENUM(SOUP_MEMORY_COPY), "SOUP_MEMORY_COPY", "copy" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupMemoryUse", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "soup-server.h" */
GType
soup_server_listen_options_get_type (void)
{
  static gsize id = 0;
  static const GFlagsValue values[] = {
    { C_FLAGS(SOUP_SERVER_LISTEN_HTTPS), "SOUP_SERVER_LISTEN_HTTPS", "https" },
    { C_FLAGS(SOUP_SERVER_LISTEN_IPV4_ONLY), "SOUP_SERVER_LISTEN_IPV4_ONLY", "ipv4-only" },
    { C_FLAGS(SOUP_SERVER_LISTEN_IPV6_ONLY), "SOUP_SERVER_LISTEN_IPV6_ONLY", "ipv6-only" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_flags_register_static ("SoupServerListenOptions", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "soup-date-utils.h" */
GType
soup_date_format_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_DATE_HTTP), "SOUP_DATE_HTTP", "http" },
    { C_ENUM(SOUP_DATE_COOKIE), "SOUP_DATE_COOKIE", "cookie" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupDateFormat", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "soup-logger.h" */
GType
soup_logger_log_level_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_LOGGER_LOG_NONE), "SOUP_LOGGER_LOG_NONE", "none" },
    { C_ENUM(SOUP_LOGGER_LOG_MINIMAL), "SOUP_LOGGER_LOG_MINIMAL", "minimal" },
    { C_ENUM(SOUP_LOGGER_LOG_HEADERS), "SOUP_LOGGER_LOG_HEADERS", "headers" },
    { C_ENUM(SOUP_LOGGER_LOG_BODY), "SOUP_LOGGER_LOG_BODY", "body" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupLoggerLogLevel", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "soup-message-headers.h" */
GType
soup_message_headers_type_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_MESSAGE_HEADERS_REQUEST), "SOUP_MESSAGE_HEADERS_REQUEST", "request" },
    { C_ENUM(SOUP_MESSAGE_HEADERS_RESPONSE), "SOUP_MESSAGE_HEADERS_RESPONSE", "response" },
    { C_ENUM(SOUP_MESSAGE_HEADERS_MULTIPART), "SOUP_MESSAGE_HEADERS_MULTIPART", "multipart" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupMessageHeadersType", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
GType
soup_encoding_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_ENCODING_UNRECOGNIZED), "SOUP_ENCODING_UNRECOGNIZED", "unrecognized" },
    { C_ENUM(SOUP_ENCODING_NONE), "SOUP_ENCODING_NONE", "none" },
    { C_ENUM(SOUP_ENCODING_CONTENT_LENGTH), "SOUP_ENCODING_CONTENT_LENGTH", "content-length" },
    { C_ENUM(SOUP_ENCODING_EOF), "SOUP_ENCODING_EOF", "eof" },
    { C_ENUM(SOUP_ENCODING_CHUNKED), "SOUP_ENCODING_CHUNKED", "chunked" },
    { C_ENUM(SOUP_ENCODING_BYTERANGES), "SOUP_ENCODING_BYTERANGES", "byteranges" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupEncoding", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
GType
soup_expectation_get_type (void)
{
  static gsize id = 0;
  static const GFlagsValue values[] = {
    { C_FLAGS(SOUP_EXPECTATION_UNRECOGNIZED), "SOUP_EXPECTATION_UNRECOGNIZED", "unrecognized" },
    { C_FLAGS(SOUP_EXPECTATION_CONTINUE), "SOUP_EXPECTATION_CONTINUE", "continue" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_flags_register_static ("SoupExpectation", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "soup-message.h" */
GType
soup_message_flags_get_type (void)
{
  static gsize id = 0;
  static const GFlagsValue values[] = {
    { C_FLAGS(SOUP_MESSAGE_NO_REDIRECT), "SOUP_MESSAGE_NO_REDIRECT", "no-redirect" },
    { C_FLAGS(SOUP_MESSAGE_NEW_CONNECTION), "SOUP_MESSAGE_NEW_CONNECTION", "new-connection" },
    { C_FLAGS(SOUP_MESSAGE_IDEMPOTENT), "SOUP_MESSAGE_IDEMPOTENT", "idempotent" },
    { C_FLAGS(SOUP_MESSAGE_DO_NOT_USE_AUTH_CACHE), "SOUP_MESSAGE_DO_NOT_USE_AUTH_CACHE", "do-not-use-auth-cache" },
    { C_FLAGS(SOUP_MESSAGE_COLLECT_METRICS), "SOUP_MESSAGE_COLLECT_METRICS", "collect-metrics" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_flags_register_static ("SoupMessageFlags", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
GType
soup_message_priority_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_MESSAGE_PRIORITY_VERY_LOW), "SOUP_MESSAGE_PRIORITY_VERY_LOW", "very-low" },
    { C_ENUM(SOUP_MESSAGE_PRIORITY_LOW), "SOUP_MESSAGE_PRIORITY_LOW", "low" },
    { C_ENUM(SOUP_MESSAGE_PRIORITY_NORMAL), "SOUP_MESSAGE_PRIORITY_NORMAL", "normal" },
    { C_ENUM(SOUP_MESSAGE_PRIORITY_HIGH), "SOUP_MESSAGE_PRIORITY_HIGH", "high" },
    { C_ENUM(SOUP_MESSAGE_PRIORITY_VERY_HIGH), "SOUP_MESSAGE_PRIORITY_VERY_HIGH", "very-high" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupMessagePriority", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "soup-session.h" */
GType
soup_session_error_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_SESSION_ERROR_PARSING), "SOUP_SESSION_ERROR_PARSING", "parsing" },
    { C_ENUM(SOUP_SESSION_ERROR_ENCODING), "SOUP_SESSION_ERROR_ENCODING", "encoding" },
    { C_ENUM(SOUP_SESSION_ERROR_TOO_MANY_REDIRECTS), "SOUP_SESSION_ERROR_TOO_MANY_REDIRECTS", "too-many-redirects" },
    { C_ENUM(SOUP_SESSION_ERROR_TOO_MANY_RESTARTS), "SOUP_SESSION_ERROR_TOO_MANY_RESTARTS", "too-many-restarts" },
    { C_ENUM(SOUP_SESSION_ERROR_REDIRECT_NO_LOCATION), "SOUP_SESSION_ERROR_REDIRECT_NO_LOCATION", "redirect-no-location" },
    { C_ENUM(SOUP_SESSION_ERROR_REDIRECT_BAD_URI), "SOUP_SESSION_ERROR_REDIRECT_BAD_URI", "redirect-bad-uri" },
    { C_ENUM(SOUP_SESSION_ERROR_MESSAGE_ALREADY_IN_QUEUE), "SOUP_SESSION_ERROR_MESSAGE_ALREADY_IN_QUEUE", "message-already-in-queue" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupSessionError", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "soup-status.h" */
GType
soup_status_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_STATUS_NONE), "SOUP_STATUS_NONE", "none" },
    { C_ENUM(SOUP_STATUS_CONTINUE), "SOUP_STATUS_CONTINUE", "continue" },
    { C_ENUM(SOUP_STATUS_SWITCHING_PROTOCOLS), "SOUP_STATUS_SWITCHING_PROTOCOLS", "switching-protocols" },
    { C_ENUM(SOUP_STATUS_PROCESSING), "SOUP_STATUS_PROCESSING", "processing" },
    { C_ENUM(SOUP_STATUS_OK), "SOUP_STATUS_OK", "ok" },
    { C_ENUM(SOUP_STATUS_CREATED), "SOUP_STATUS_CREATED", "created" },
    { C_ENUM(SOUP_STATUS_ACCEPTED), "SOUP_STATUS_ACCEPTED", "accepted" },
    { C_ENUM(SOUP_STATUS_NON_AUTHORITATIVE), "SOUP_STATUS_NON_AUTHORITATIVE", "non-authoritative" },
    { C_ENUM(SOUP_STATUS_NO_CONTENT), "SOUP_STATUS_NO_CONTENT", "no-content" },
    { C_ENUM(SOUP_STATUS_RESET_CONTENT), "SOUP_STATUS_RESET_CONTENT", "reset-content" },
    { C_ENUM(SOUP_STATUS_PARTIAL_CONTENT), "SOUP_STATUS_PARTIAL_CONTENT", "partial-content" },
    { C_ENUM(SOUP_STATUS_MULTI_STATUS), "SOUP_STATUS_MULTI_STATUS", "multi-status" },
    { C_ENUM(SOUP_STATUS_MULTIPLE_CHOICES), "SOUP_STATUS_MULTIPLE_CHOICES", "multiple-choices" },
    { C_ENUM(SOUP_STATUS_MOVED_PERMANENTLY), "SOUP_STATUS_MOVED_PERMANENTLY", "moved-permanently" },
    { C_ENUM(SOUP_STATUS_FOUND), "SOUP_STATUS_FOUND", "found" },
    { C_ENUM(SOUP_STATUS_MOVED_TEMPORARILY), "SOUP_STATUS_MOVED_TEMPORARILY", "moved-temporarily" },
    { C_ENUM(SOUP_STATUS_SEE_OTHER), "SOUP_STATUS_SEE_OTHER", "see-other" },
    { C_ENUM(SOUP_STATUS_NOT_MODIFIED), "SOUP_STATUS_NOT_MODIFIED", "not-modified" },
    { C_ENUM(SOUP_STATUS_USE_PROXY), "SOUP_STATUS_USE_PROXY", "use-proxy" },
    { C_ENUM(SOUP_STATUS_NOT_APPEARING_IN_THIS_PROTOCOL), "SOUP_STATUS_NOT_APPEARING_IN_THIS_PROTOCOL", "not-appearing-in-this-protocol" },
    { C_ENUM(SOUP_STATUS_TEMPORARY_REDIRECT), "SOUP_STATUS_TEMPORARY_REDIRECT", "temporary-redirect" },
    { C_ENUM(SOUP_STATUS_PERMANENT_REDIRECT), "SOUP_STATUS_PERMANENT_REDIRECT", "permanent-redirect" },
    { C_ENUM(SOUP_STATUS_BAD_REQUEST), "SOUP_STATUS_BAD_REQUEST", "bad-request" },
    { C_ENUM(SOUP_STATUS_UNAUTHORIZED), "SOUP_STATUS_UNAUTHORIZED", "unauthorized" },
    { C_ENUM(SOUP_STATUS_PAYMENT_REQUIRED), "SOUP_STATUS_PAYMENT_REQUIRED", "payment-required" },
    { C_ENUM(SOUP_STATUS_FORBIDDEN), "SOUP_STATUS_FORBIDDEN", "forbidden" },
    { C_ENUM(SOUP_STATUS_NOT_FOUND), "SOUP_STATUS_NOT_FOUND", "not-found" },
    { C_ENUM(SOUP_STATUS_METHOD_NOT_ALLOWED), "SOUP_STATUS_METHOD_NOT_ALLOWED", "method-not-allowed" },
    { C_ENUM(SOUP_STATUS_NOT_ACCEPTABLE), "SOUP_STATUS_NOT_ACCEPTABLE", "not-acceptable" },
    { C_ENUM(SOUP_STATUS_PROXY_AUTHENTICATION_REQUIRED), "SOUP_STATUS_PROXY_AUTHENTICATION_REQUIRED", "proxy-authentication-required" },
    { C_ENUM(SOUP_STATUS_PROXY_UNAUTHORIZED), "SOUP_STATUS_PROXY_UNAUTHORIZED", "proxy-unauthorized" },
    { C_ENUM(SOUP_STATUS_REQUEST_TIMEOUT), "SOUP_STATUS_REQUEST_TIMEOUT", "request-timeout" },
    { C_ENUM(SOUP_STATUS_CONFLICT), "SOUP_STATUS_CONFLICT", "conflict" },
    { C_ENUM(SOUP_STATUS_GONE), "SOUP_STATUS_GONE", "gone" },
    { C_ENUM(SOUP_STATUS_LENGTH_REQUIRED), "SOUP_STATUS_LENGTH_REQUIRED", "length-required" },
    { C_ENUM(SOUP_STATUS_PRECONDITION_FAILED), "SOUP_STATUS_PRECONDITION_FAILED", "precondition-failed" },
    { C_ENUM(SOUP_STATUS_REQUEST_ENTITY_TOO_LARGE), "SOUP_STATUS_REQUEST_ENTITY_TOO_LARGE", "request-entity-too-large" },
    { C_ENUM(SOUP_STATUS_REQUEST_URI_TOO_LONG), "SOUP_STATUS_REQUEST_URI_TOO_LONG", "request-uri-too-long" },
    { C_ENUM(SOUP_STATUS_UNSUPPORTED_MEDIA_TYPE), "SOUP_STATUS_UNSUPPORTED_MEDIA_TYPE", "unsupported-media-type" },
    { C_ENUM(SOUP_STATUS_REQUESTED_RANGE_NOT_SATISFIABLE), "SOUP_STATUS_REQUESTED_RANGE_NOT_SATISFIABLE", "requested-range-not-satisfiable" },
    { C_ENUM(SOUP_STATUS_INVALID_RANGE), "SOUP_STATUS_INVALID_RANGE", "invalid-range" },
    { C_ENUM(SOUP_STATUS_EXPECTATION_FAILED), "SOUP_STATUS_EXPECTATION_FAILED", "expectation-failed" },
    { C_ENUM(SOUP_STATUS_MISDIRECTED_REQUEST), "SOUP_STATUS_MISDIRECTED_REQUEST", "misdirected-request" },
    { C_ENUM(SOUP_STATUS_UNPROCESSABLE_ENTITY), "SOUP_STATUS_UNPROCESSABLE_ENTITY", "unprocessable-entity" },
    { C_ENUM(SOUP_STATUS_LOCKED), "SOUP_STATUS_LOCKED", "locked" },
    { C_ENUM(SOUP_STATUS_FAILED_DEPENDENCY), "SOUP_STATUS_FAILED_DEPENDENCY", "failed-dependency" },
    { C_ENUM(SOUP_STATUS_INTERNAL_SERVER_ERROR), "SOUP_STATUS_INTERNAL_SERVER_ERROR", "internal-server-error" },
    { C_ENUM(SOUP_STATUS_NOT_IMPLEMENTED), "SOUP_STATUS_NOT_IMPLEMENTED", "not-implemented" },
    { C_ENUM(SOUP_STATUS_BAD_GATEWAY), "SOUP_STATUS_BAD_GATEWAY", "bad-gateway" },
    { C_ENUM(SOUP_STATUS_SERVICE_UNAVAILABLE), "SOUP_STATUS_SERVICE_UNAVAILABLE", "service-unavailable" },
    { C_ENUM(SOUP_STATUS_GATEWAY_TIMEOUT), "SOUP_STATUS_GATEWAY_TIMEOUT", "gateway-timeout" },
    { C_ENUM(SOUP_STATUS_HTTP_VERSION_NOT_SUPPORTED), "SOUP_STATUS_HTTP_VERSION_NOT_SUPPORTED", "http-version-not-supported" },
    { C_ENUM(SOUP_STATUS_INSUFFICIENT_STORAGE), "SOUP_STATUS_INSUFFICIENT_STORAGE", "insufficient-storage" },
    { C_ENUM(SOUP_STATUS_NOT_EXTENDED), "SOUP_STATUS_NOT_EXTENDED", "not-extended" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupStatus", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
GType
soup_http_version_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_HTTP_1_0), "SOUP_HTTP_1_0", "http-1-0" },
    { C_ENUM(SOUP_HTTP_1_1), "SOUP_HTTP_1_1", "http-1-1" },
    { C_ENUM(SOUP_HTTP_2_0), "SOUP_HTTP_2_0", "http-2-0" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupHTTPVersion", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "soup-tld.h" */
GType
soup_tld_error_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_TLD_ERROR_INVALID_HOSTNAME), "SOUP_TLD_ERROR_INVALID_HOSTNAME", "invalid-hostname" },
    { C_ENUM(SOUP_TLD_ERROR_IS_IP_ADDRESS), "SOUP_TLD_ERROR_IS_IP_ADDRESS", "is-ip-address" },
    { C_ENUM(SOUP_TLD_ERROR_NOT_ENOUGH_DOMAINS), "SOUP_TLD_ERROR_NOT_ENOUGH_DOMAINS", "not-enough-domains" },
    { C_ENUM(SOUP_TLD_ERROR_NO_BASE_DOMAIN), "SOUP_TLD_ERROR_NO_BASE_DOMAIN", "no-base-domain" },
    { C_ENUM(SOUP_TLD_ERROR_NO_PSL_DATA), "SOUP_TLD_ERROR_NO_PSL_DATA", "no-psl-data" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupTLDError", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "soup-uri-utils.h" */
GType
soup_uri_component_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_URI_NONE), "SOUP_URI_NONE", "none" },
    { C_ENUM(SOUP_URI_SCHEME), "SOUP_URI_SCHEME", "scheme" },
    { C_ENUM(SOUP_URI_USER), "SOUP_URI_USER", "user" },
    { C_ENUM(SOUP_URI_PASSWORD), "SOUP_URI_PASSWORD", "password" },
    { C_ENUM(SOUP_URI_AUTH_PARAMS), "SOUP_URI_AUTH_PARAMS", "auth-params" },
    { C_ENUM(SOUP_URI_HOST), "SOUP_URI_HOST", "host" },
    { C_ENUM(SOUP_URI_PORT), "SOUP_URI_PORT", "port" },
    { C_ENUM(SOUP_URI_PATH), "SOUP_URI_PATH", "path" },
    { C_ENUM(SOUP_URI_QUERY), "SOUP_URI_QUERY", "query" },
    { C_ENUM(SOUP_URI_FRAGMENT), "SOUP_URI_FRAGMENT", "fragment" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupURIComponent", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
/* enumerations from "soup-websocket.h" */
GType
soup_websocket_error_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_WEBSOCKET_ERROR_FAILED), "SOUP_WEBSOCKET_ERROR_FAILED", "failed" },
    { C_ENUM(SOUP_WEBSOCKET_ERROR_NOT_WEBSOCKET), "SOUP_WEBSOCKET_ERROR_NOT_WEBSOCKET", "not-websocket" },
    { C_ENUM(SOUP_WEBSOCKET_ERROR_BAD_HANDSHAKE), "SOUP_WEBSOCKET_ERROR_BAD_HANDSHAKE", "bad-handshake" },
    { C_ENUM(SOUP_WEBSOCKET_ERROR_BAD_ORIGIN), "SOUP_WEBSOCKET_ERROR_BAD_ORIGIN", "bad-origin" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupWebsocketError", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
GType
soup_websocket_connection_type_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_WEBSOCKET_CONNECTION_UNKNOWN), "SOUP_WEBSOCKET_CONNECTION_UNKNOWN", "unknown" },
    { C_ENUM(SOUP_WEBSOCKET_CONNECTION_CLIENT), "SOUP_WEBSOCKET_CONNECTION_CLIENT", "client" },
    { C_ENUM(SOUP_WEBSOCKET_CONNECTION_SERVER), "SOUP_WEBSOCKET_CONNECTION_SERVER", "server" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupWebsocketConnectionType", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
GType
soup_websocket_data_type_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_WEBSOCKET_DATA_TEXT), "SOUP_WEBSOCKET_DATA_TEXT", "text" },
    { C_ENUM(SOUP_WEBSOCKET_DATA_BINARY), "SOUP_WEBSOCKET_DATA_BINARY", "binary" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupWebsocketDataType", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
GType
soup_websocket_close_code_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_WEBSOCKET_CLOSE_NORMAL), "SOUP_WEBSOCKET_CLOSE_NORMAL", "normal" },
    { C_ENUM(SOUP_WEBSOCKET_CLOSE_GOING_AWAY), "SOUP_WEBSOCKET_CLOSE_GOING_AWAY", "going-away" },
    { C_ENUM(SOUP_WEBSOCKET_CLOSE_PROTOCOL_ERROR), "SOUP_WEBSOCKET_CLOSE_PROTOCOL_ERROR", "protocol-error" },
    { C_ENUM(SOUP_WEBSOCKET_CLOSE_UNSUPPORTED_DATA), "SOUP_WEBSOCKET_CLOSE_UNSUPPORTED_DATA", "unsupported-data" },
    { C_ENUM(SOUP_WEBSOCKET_CLOSE_NO_STATUS), "SOUP_WEBSOCKET_CLOSE_NO_STATUS", "no-status" },
    { C_ENUM(SOUP_WEBSOCKET_CLOSE_ABNORMAL), "SOUP_WEBSOCKET_CLOSE_ABNORMAL", "abnormal" },
    { C_ENUM(SOUP_WEBSOCKET_CLOSE_BAD_DATA), "SOUP_WEBSOCKET_CLOSE_BAD_DATA", "bad-data" },
    { C_ENUM(SOUP_WEBSOCKET_CLOSE_POLICY_VIOLATION), "SOUP_WEBSOCKET_CLOSE_POLICY_VIOLATION", "policy-violation" },
    { C_ENUM(SOUP_WEBSOCKET_CLOSE_TOO_BIG), "SOUP_WEBSOCKET_CLOSE_TOO_BIG", "too-big" },
    { C_ENUM(SOUP_WEBSOCKET_CLOSE_NO_EXTENSION), "SOUP_WEBSOCKET_CLOSE_NO_EXTENSION", "no-extension" },
    { C_ENUM(SOUP_WEBSOCKET_CLOSE_SERVER_ERROR), "SOUP_WEBSOCKET_CLOSE_SERVER_ERROR", "server-error" },
    { C_ENUM(SOUP_WEBSOCKET_CLOSE_TLS_HANDSHAKE), "SOUP_WEBSOCKET_CLOSE_TLS_HANDSHAKE", "tls-handshake" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupWebsocketCloseCode", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}
GType
soup_websocket_state_get_type (void)
{
  static gsize id = 0;
  static const GEnumValue values[] = {
    { C_ENUM(SOUP_WEBSOCKET_STATE_OPEN), "SOUP_WEBSOCKET_STATE_OPEN", "open" },
    { C_ENUM(SOUP_WEBSOCKET_STATE_CLOSING), "SOUP_WEBSOCKET_STATE_CLOSING", "closing" },
    { C_ENUM(SOUP_WEBSOCKET_STATE_CLOSED), "SOUP_WEBSOCKET_STATE_CLOSED", "closed" },
    { 0, NULL, NULL }
  };

  if (g_once_init_enter (&id)) {
    GType tmp = g_enum_register_static ("SoupWebsocketState", values);
    g_once_init_leave (&id, tmp);
  }

  return (GType) id;
}


/* Generated data ends here */

